"""
This file was auto-generated by djelm {{cookiecutter.version}}

For the best developer experience ensure the following:

- {{cookiecutter.program_name}}Flags variable is not renamed.
- The key variable value "{{cookiecutter.scope}}-djelmwidget-{{cookiecutter.program_name}}" is not modified
"""

from djelm.flags import Flags
from djelm.flags.form import ModelMultipleChoiceFieldFlag

key = "{{cookiecutter.scope}}-djelmwidget-{{cookiecutter.view_name}}"

{{cookiecutter.program_name}}Flags = Flags(ModelMultipleChoiceFieldFlag())
